﻿--[[
	MrtWoo_AntiSpam: Detects and blocks russian commercial spam and flud
	Copyright (C) 2010 Pavel Dudkovsky (mrtime@era.by)
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
]]

local L = LibStub("AceLocale-3.0"):GetLocale("MrtWoo_AntiSpam", false);

-- ########################################################################

function __genOrderedIndex( t )
	local orderedIndex = {}
	for key in pairs(t) do
		table.insert( orderedIndex, key )
	end
	table.sort( orderedIndex )
	return orderedIndex
end

function orderedNext(t, state)
	-- Equivalent of the next function, but returns the keys in the alphabetic
	-- order. We use a temporary ordered key table that is stored in the
	-- table being iterated.

	--print("orderedNext: state = "..tostring(state) )
	if state == nil then
		-- the first time, generate the index
		t.__orderedIndex = __genOrderedIndex( t )
		key = t.__orderedIndex[1]
		return key, t[key]
	end
	-- fetch the next value
	key = nil
	for i = 1,table.getn(t.__orderedIndex) do
		if t.__orderedIndex[i] == state then
			key = t.__orderedIndex[i+1]
		end
	end

	if key then
		return key, t[key]
	end

	-- no more value to return, cleanup
	t.__orderedIndex = nil
	return
end

function orderedPairs(t)
	-- Equivalent of the pairs() function on tables. Allows to iterate
	-- in order
	return orderedNext, t, nil
end

-- ########################################################################

MrtWoo_AntiSpamMonitor = {};
-- MrtWoo_AntiSpamMonitor.UInfo = {};

MrtWoo_AntiSpamMonitor.MSG_RESULT_OK = 0;
MrtWoo_AntiSpamMonitor.MSG_RESULT_SPAM = 1;
MrtWoo_AntiSpamMonitor.MSG_RESULT_FLUD = 2;
MrtWoo_AntiSpamMonitor.MSG_RESULT_WAIT = 3;

MrtWoo_AntiSpamMonitor.MSG_STATUS_WAIT_PROCESSING = 0;
MrtWoo_AntiSpamMonitor.MSG_STATUS_DONE = 100;

MrtWoo_AntiSpamMonitor.SMI_COUNT_OK = 0;
MrtWoo_AntiSpamMonitor.SMI_COUNT_FLUD = 0;
MrtWoo_AntiSpamMonitor.SMI_COUNT_SPAM = 0;
MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_ADD = 0;
MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_WAIT = 0;
MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST = 0;
MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_MAX = 0;

MrtWoo_AntiSpamMonitor.Players = setmetatable({},{__mode="k"});

function MrtWoo_AntiSpamMonitor:new(o)
	o = o or {};
	setmetatable(o, self);
	self.__index = self;

	LibStub:GetLibrary('AceEvent-3.0'):Embed(self);
	
	o.Players = setmetatable({},{__mode="k"});
	o.MessagesQueue = setmetatable({},{__mode="k"});	
	o.MessagesFilter = MrtWoo_AntiSpamBlocFilter:new();
	o.Who = LibStub:GetLibrary("LibWho-2.0");	
	o.Who:RegisterCallback('WHOLIB_QUERY_RESULT', MrtWoo_AntiSpamMonitor.WhoLibCallback);

	self.Who = o.Who;
	self.Players = o.Players;
	self.MessagesFilter = o.MessagesFilter;
	self.MessagesQueue = o.MessagesQueue;
	
	return o;
end

function MrtWoo_AntiSpamMonitor:ShowSMI()
	MrtWoo:Print(
		  'QA:', MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_ADD, 
		', QW:', MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_WAIT,
		', QL:', MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST,
		', QM:', MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_MAX,
		', SO:', MrtWoo_AntiSpamMonitor.SMI_COUNT_OK,
		', SF:', MrtWoo_AntiSpamMonitor.SMI_COUNT_FLUD,
		', SS:', MrtWoo_AntiSpamMonitor.SMI_COUNT_SPAM
	);		
end;

MrtWoo_AntiSpamMonitor.OldSMI = {0,0,0,0,0,0,0};

function MrtWoo_AntiSpamMonitor:SendSMI()
	local o = MrtWoo_AntiSpamMonitor.OldSMI;
	local v = {
		MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_ADD, 
		MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_WAIT,
		MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST,
		MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_MAX,
		MrtWoo_AntiSpamMonitor.SMI_COUNT_OK,
		MrtWoo_AntiSpamMonitor.SMI_COUNT_FLUD,
		MrtWoo_AntiSpamMonitor.SMI_COUNT_SPAM	
	};
	
	if 	o[1] ~= v[1] or o[2] ~= v[2] or o[3] ~= v[3] or o[4] ~= v[4] or 
		o[5] ~= v[5] or o[6] ~= v[6] or o[7] ~= v[7] then			
			self:SendMessage('MRTWOO_ASMON_SMI', v[1], v[2], v[3], v[4], v[5], v[6], v[7]);
			MrtWoo_AntiSpamMonitor.OldSMI = v;
	end;
end;

function MrtWoo_AntiSpamMonitor:AddMessageToQueue(msgId, ...)
	if self:GetMessage(msgId) then
		return false;
	end;
	
	MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_ADD = MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_ADD + 1;	
	
	local v = {};
	v.time = GetTime();
	v.status = MrtWoo_AntiSpamMonitor.MSG_STATUS_WAIT_PROCESSING;
	v.data = {...};
	
	self.MessagesQueue[msgId] = v ;
	self:ProcessQueue();

	return true;
end;

function MrtWoo_AntiSpamMonitor:GetMessage(msgId)
	return self.MessagesQueue[msgId];
end;

function MrtWoo_AntiSpamMonitor:ProcessQueue()
	--[[
	if MrtWoo.db.profile.modules.as_debug then
		MrtWoo:Print('=== MrtWoo_AntiSpamMonitor:ProcessQueue(): I ===');
	end;
	]]
	
	if (self.MessagesQueue == nil) then
		return false;
	end;
	
	local c_time = GetTime();
	
	MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_WAIT = 0;
	MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST = 0;
	
	for k, v in orderedPairs(self.MessagesQueue) do
		if v.status ~= MrtWoo_AntiSpamMonitor.MSG_STATUS_DONE then
		
			local message = v.data[2];
			local author = v.data[3];
			local msgId = v.data[12];
			
			local msgStatus = MrtWoo_AntiSpamMonitor.MSG_RESULT_OK;
			
			if MrtWoo.MyName ~= author then
				msgStatus = self:CheckMessage(msgId, message, author);
			end;
			
			if msgStatus == MrtWoo_AntiSpamMonitor.MSG_RESULT_OK then 
				self:SendMessage('MRTWOO_ASMON_PROCESSED_MESSAGE', 
					'CHANNEL', 
					v.data[1],  v.data[2],  v.data[3],  v.data[4],  v.data[5], 
					v.data[6],  v.data[7],  v.data[8],  v.data[9],  v.data[10], 
					v.data[11], v.data[12], v.data[13], v.data[14],	v.data[15], 
					v.data[16], v.data[17], v.data[18], v.data[19], v.data[20]					
				);
				MrtWoo_AntiSpamMonitor.SMI_COUNT_OK = MrtWoo_AntiSpamMonitor.SMI_COUNT_OK + 1;
			end;
			
			if msgStatus ~= MrtWoo_AntiSpamMonitor.MSG_RESULT_WAIT then					
			
				if msgStatus == MrtWoo_AntiSpamMonitor.MSG_RESULT_SPAM then	
					MrtWoo_AntiSpamMonitor.SMI_COUNT_SPAM = MrtWoo_AntiSpamMonitor.SMI_COUNT_SPAM + 1;
				elseif msgStatus == MrtWoo_AntiSpamMonitor.MSG_RESULT_FLUD then	
					MrtWoo_AntiSpamMonitor.SMI_COUNT_FLUD = MrtWoo_AntiSpamMonitor.SMI_COUNT_FLUD + 1;
				end;
				
				v.time = GetTime();
				v.status = MrtWoo_AntiSpamMonitor.MSG_STATUS_DONE;
				v.data = nil;
				
				self.MessagesQueue[k] = v;
			else
				MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_WAIT = MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_WAIT + 1;
			end;
		else
			if c_time - v.time > MrtWoo.db.profile.modules.as_gc_message_lifetime then 
				--[[
				if MrtWoo.db.profile.modules.as_debug then
					MrtWoo:Print('S REMOVE', k, c_time - v.time);
				end;
				]]
				self.MessagesQueue[k] = nil;
			end;
					
		end;
		
		MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST = MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST + 1;
	end
		
	if MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_MAX < MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST then
		MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_MAX = MrtWoo_AntiSpamMonitor.SMI_COUNT_QUEUE_LAST;
	end;
	
	self:SendSMI();

	--[[
	if MrtWoo.db.profile.modules.as_debug then	
		MrtWoo:Print('=== MrtWoo_AntiSpamMonitor:ProcessQueue(): O ===');	
	end;
	]]
	
	return true;
end;

function MrtWoo_AntiSpamMonitor:WhoLibCallback(query, results, complete)
	if not complete then
		if MrtWoo.db.profile.modules.as_debug then		
			MrtWoo:Print('There were more Players...');
		end;
	else
		for _,result in pairs(results) do
			if MrtWoo.db.profile.modules.as_debug then		
				MrtWoo:Print('Player: ' .. result.Name .. ' , level: ' .. result.Level);
			end;
			local lwrName = string.lower(result.Name);
			-- MrtWoo_AntiSpamMonitor.UInfo[lwrName] = result;

			-- NEW
			local player = MrtWoo_AntiSpamMonitor.Players[lwrName];	
			
			if not player then
				player = MrtWoo_AntiSpamPlayer:new();
				player:SetName(msgAuthor);
				MrtWoo_AntiSpamMonitor.Players[lwrName] = player;
			end;

			player:SetInfo(result);
			player:UnLock();
						
		end;
	end;
	
	MrtWoo_AntiSpamMonitor:ProcessQueue();
end;

function MrtWoo_AntiSpamMonitor:GetUserInfo(player)

	local usrInfo = player:GetInfo();
	
	if usrInfo == nil then
		
		player:SetInfo(false);
		
		local options = {
			queue = self.Who.WHOLIB_QUERY_QUIET,
		};
		
		self.Who:Who(string.format('и-"%s"', player:GetName()), options);
		
	elseif type(usrInfo) == 'table' then
		
		return usrInfo;

	end;
	
	return false;	
end;

function MrtWoo_AntiSpamMonitor:RateUserInfo(player)
	local userInfo = MrtWoo_AntiSpamMonitor:GetUserInfo(player);
	
	if not userInfo then
		return nil;
	end;
	
	if userInfo.Level <= MrtWoo.db.profile.modules.as_spam_maxlvl then
		return MrtWoo.db.profile.modules.as_spam_ratelvl;
	end;
	
	return 0;
end;

function MrtWoo_AntiSpamMonitor:CheckMessage(msgId, msgVal, msgAuthor)
	local lwrName = string.lower(msgAuthor);
	local player = self.Players[lwrName];

	local result = self.MSG_RESULT_OK;
	
	if not player then
		player = MrtWoo_AntiSpamPlayer:new();
		player:SetName(msgAuthor);
		
		self.Players[lwrName] = player;
		
		-- MrtWoo:Print('NEW:', msgAuthor);
	end;
	
	local messages = player:GetMessages();
	
	local msgRate = 0;
			
	if not player:IsSpammer() and not player:IsLock() then

		-- MrtWoo:Print(msgAuthor, messages);
		
		local _msg = messages:GetMsg(msgId);
		
		if not _msg then
		
			if MrtWoo.db.profile.modules.as_antiflud then
				local _dup = messages:GetDup(msgVal);

				if _dup then
					local _time = GetTime();
					
					if ((string.len(msgVal) >= 10) and (_time - _dup.t) < MrtWoo.db.profile.modules.as_antiflud_time) then
						if MrtWoo.db.profile.modules.as_debug then
							MrtWoo:Print(string.format("%s, %g - BLOCK (FLUD)", msgAuthor, (_time - _dup.t)));
						end;
						result = self.MSG_RESULT_FLUD;
					else
						if MrtWoo.db.profile.modules.as_debug then
							MrtWoo:Print(string.format("%s, %g - ALLOW", msgAuthor, (_time - _dup.t)));
						end;
					end;
				end;
			end;
		
			_msg = messages:Add(msgId, msgVal);
			
			if (result == self.MSG_RESULT_OK) then
				msgRate = self.MessagesFilter:RateTradeMessage(msgVal);
			end;
		else
			msgRate = _msg:GetRate();
		end;
	
		if (result == self.MSG_RESULT_OK) then
			if MrtWoo.db.profile.modules.as_spam_additional_checks and not player:IsLock() then
				-- if true then
				if  (msgRate >= MrtWoo.db.profile.modules.as_spamrate_needed_for_additional_checks) and 
					(msgRate < MrtWoo.db.profile.modules.as_spamrate) then

					local userRate = MrtWoo_AntiSpamMonitor:RateUserInfo(player);
					if userRate == nil then
						if MrtWoo.db.profile.modules.as_debug then
							MrtWoo:Print(string.format(string.format("REQ_INFO: msgAuthor: %s", msgAuthor)));
						end;
						player:Lock();
					else
						msgRate = msgRate + userRate;
					end;
				end;
			end;
					
			if (msgRate >= MrtWoo.db.profile.modules.as_spamrate) then
				player:Spammer();
				
				if not MrtWoo.db.profile.modules.as_silent or MrtWoo.db.profile.modules.as_debug then
					MrtWoo:Print(string.format(L["Messages from the player %s locked, reason: Spam! Post rating: %i"], msgAuthor, msgRate));
				end;
				
				if MrtWoo.db.profile.modules.as_report and not _msg:IsReported() then
					_msg:Report();
				end;				
			end;	
		end;
	end;
	
	if player:IsSpammer() then
		result = self.MSG_RESULT_SPAM;
	elseif player:IsLock() then 
		result = self.MSG_RESULT_WAIT;
	end;
		
	if MrtWoo.db.profile.modules.as_debug and (result == self.MSG_RESULT_SPAM or result == self.MSG_RESULT_FLUD) then
		MrtWoo:Print(string.format(string.format("!OK: msgId: %i, msgVal: %s, msgRate: %i, msgAuthor: %s", msgId, msgVal, msgRate, msgAuthor)));
	end;
	
	if 	result == self.MSG_RESULT_OK then
		local _dup = messages:GetDup(msgVal);
		_dup.t = GetTime();
	end;

	
	return result;	
	
end;

--[[
function MrtWoo_AntiSpamMonitor:_CheckMessage(msgId, message, username)
	
	self:SmartClean();
	
	local lwrName = string.lower(username);
	local res = self.Players[lwrName];
	
	
	local v = {};
	v.rate = 0;
	v.time = GetTime();
	v.time_last_ok_msg = 0;
	v.is_spam = false;
	v.is_flud = false;
	v.is_lock = false;

	if not res then
		v.message = message;
		v.rate = self.MessagesFilter:RateTradeMessage(message);
	else
		
		v.time_last_ok_msg = res.time_last_ok_msg;
							
		if not res.is_lock then						
			v.message = message;
			
			if not res.is_spam then			
				if (v.message ~= res.message) or not MrtWoo.db.profile.modules.as_antiflud then
					v.rate = self.MessagesFilter:RateTradeMessage(v.message);
				else
					if (v.time_last_ok_msg > 0) and ((v.time - v.time_last_ok_msg) < MrtWoo.db.profile.modules.as_antiflud_time) then
						if MrtWoo.db.profile.modules.as_debug then
							MrtWoo:Print(string.format("%s, %g - BLOCK (FLUD)", username, (v.time - v.time_last_ok_msg)));
						end;
						v.is_flud = true;
					else
						if MrtWoo.db.profile.modules.as_debug then
							MrtWoo:Print(string.format("%s, %g - ALLOW", username, (v.time - v.time_last_ok_msg)));
						end;
					end;					
				end;
			else
				v.is_spam = true;
			end;
			
		else
			v.message = res.message;
			v.rate = res.rate;
		
			if (v.message ~= message) then
				v.is_lock = true;
			end;
		end;
	end;
	
	if MrtWoo.db.profile.modules.as_spam_additional_checks and not v.is_lock then
		-- if true then
		if  (v.rate >= MrtWoo.db.profile.modules.as_spamrate_needed_for_additional_checks) and 
			(v.rate < MrtWoo.db.profile.modules.as_spamrate) then
		
			local userRate = MrtWoo_AntiSpamMonitor:RateUserInfo(username);
			if userRate == nil then
				if MrtWoo.db.profile.modules.as_debug then
					MrtWoo:Print('LOCK', username, message);
				end;
				v.is_lock = true;
			else
				if MrtWoo.db.profile.modules.as_debug then
					if res.is_lock then
						MrtWoo:Print('UNLOCK', username, ', in: ', v.rate, 'out: ', v.rate + userRate);
					else
						MrtWoo:Print(username, ', in: ', v.rate, 'out: ', v.rate + userRate);
					end;
				end;
				v.rate = v.rate + userRate;
			end;
			
		end;
	end;
	if not v.is_lock then		
		if (v.rate >= MrtWoo.db.profile.modules.as_spamrate) then
			v.is_spam = true;
			if not MrtWoo.db.profile.modules.as_silent or MrtWoo.db.profile.modules.as_debug then
				MrtWoo:Print(string.format(L["Messages from the player %s locked, reason: Spam! Post rating: %i"], username, v.rate));
			end;
		end;	
		
		if MrtWoo.db.profile.modules.as_debug and (v.is_spam or v.is_flud) then
			MrtWoo:Print('!OK', username, message);
		end;
	end;
	
	local result = self.MSG_RESULT_OK;
	
	if v.is_lock then
		result = self.MSG_RESULT_WAIT;
	elseif v.is_spam then
		result = self.MSG_RESULT_SPAM;
	elseif v.is_flud then
		result = self.MSG_RESULT_FLUD;
	end;

	if result == self.MSG_RESULT_OK then
		v.time_last_ok_msg = GetTime();
	end;
	
	self.Players[lwrName] = v;
		
	return result;
	
end;

MrtWoo_AntiSpamMonitor.LastCleanTime = 0;

function MrtWoo_AntiSpamMonitor:SmartClean()
	local c_time = GetTime();
	if (c_time - self.LastCleanTime) > MrtWoo.db.profile.modules.as_gc_wait_time then 

		table.foreach(self.Players, 
			function(k, v) 
				if not v.is_lock and (c_time - v.time) > MrtWoo.db.profile.modules.as_gc_Players_lifetime then
					if MrtWoo.db.profile.modules.as_debug then
						MrtWoo:Print('U REMOVE', k, c_time - v.time);
					end;
					self.Players[k] = nil;
				end;
			end
		);
				
		self.LastCleanTime = c_time;
	end;
end;

]]
